/*
  evexex (ver. 0.1) -- Evolutionary Explanation of the Experiments

  Copyright (C) 2006 Mikhail Anufriev

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  (version 2) as published by the Free Software Foundation;
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "tools.h"

/* Definitions of structures */
/* ========================= */

/* Rules structure */
/* --------------- */

typedef struct rule {

  void *r;             /* rule */
 
  int argnum;
  char **args;
  
  char     **argtype;  /* variable type */
  unsigned *arglag;    /* variable lag */
  double   *values;    /* variable values */

  double *pexp;        /* expected prices according to the rule */  
  double u;            /* performance of the rule */
  double n;            /* share of agents using the rule */

  unsigned t;          /* time period when the rule starts to be used */  
  
  int maxlag;       /* max lag of the rule */

} Rule;


/* Market structure */
/* ---------------- */
typedef struct market {
  
  double rf;           /* risk free interest rate */
  
  double y;            /* current dividend realization*/
  int dividend_type;   /* type of the dividend process*/
  double process_mean; /* mean of the process (dividend or dividend growth) */
  double process_stdev;/* std.dev. of the process (dividend or dividend growth)*/
  
  double pf;           /* fundamental price */
  
  double *p;           /* current and historical market price */  
  
  unsigned t;          /* progressive time index */
  
  double *pinit;       /* initial price history */
  int pinum; 
  
  double nrobot;       /* share of robot traders */
  
  double beta;         /* intencity of choice */
  double eta;          /* memory in the performance measure */
  double delta;        /* fraction of conservative traders who do update their strategy */
  
  gsl_rng *rng;  

} Market;


/* Dividend processes */
/* ------------------ */
double dividend_normal (const Market market )       /* CASE 0: normally distributed dividend */
{
  return gsl_ran_gaussian(market.rng,market.process_stdev)+market.process_mean;
}

double dividend_experiment (const Market market )   /* CASE 1: dividend is taken from the experiment */
{
  double div;
  static double exper[52] = {
    -0.763877, 0.883503, 0.23916, -0.467155, 0.576525, -0.230388, -0.648788, -0.625582, -0.597419,
    -0.109467, -0.202767, 0.107619, 0.219005, 0.0814408, 0.122126, 0.60628, 0.36566, 0.0523999, 
    -0.763833, -0.765342, 0.1704, 0.827883, -0.662045, -0.0308038, 0.224827, -0.189083, 0.250096,
    -0.203002, 0.498039, -0.570139, -0.854882, 0.219192, -0.780388, 0.631271, 0.181479, 0.37667,
    0.0995356, -0.355682, -0.296422, 0.176621, -0.774381, -0.578058, 0.911034, -0.153825, -0.267501,
    0.756582, -0.180097, 0.0600999, -0.782315, 0.166507, 0.761548, 0.0239382
  };
  if (market.t<52) 
    div = exper[market.t]+market.process_mean;
  else 
    div = gsl_ran_gaussian(market.rng,market.process_stdev)+market.process_mean;
  return div;  
}

/* Rule and market initialization */
/* =============================== */

/* define rule from the string */
/* ---------------------------- */
void rule_from_string(Rule *rule,const char *definition,const int maxtime)
{ 
  /* variables for reading suboptions */
  char *string = strdup(definition);
  char *stmp1  = string;
  char *value;
  int opt;
    
  /* possibles specifiers in the rule's description line */
  char *const rule_parameters[] = {"r","n",NULL};
  
  rule->r = NULL;       /* default: no function    */
  rule->t=1;            /* initial starting time is 0 */
  rule->u=0;            /* initial performance of the rule is 0 */
  rule->n=0;            /* initial share of the rule is 0 */
  rule->maxlag=0;       /* default value of maxlag is 0 */ 
  
  while (*string != '\0')  
    switch (opt = getsubopt(&string, rule_parameters, &value)){
    case 0:  /* functional form of rule */
      {
	int i;
	
	/* define rule function and list of arguments */
	rule->r = evaluator_create (value);
	assert(rule->r);
	evaluator_get_variables (rule->r, &(rule->args), &(rule->argnum));
	
	/* allocate the variables that describe the arguments and values */
	rule->argtype=(char **) my_alloc(rule->argnum * 3 * sizeof(char));
	rule->arglag =(unsigned *) my_alloc(rule->argnum * sizeof(unsigned));
	rule->pexp =(double *) my_alloc(maxtime * sizeof(double));
	rule->values = (double *) my_alloc(rule->argnum * sizeof(double));
	
	/* parse the rule function */ 
	for(i=0;i<rule->argnum;i++){     
	  char *type = rule->args[i];           /* set the type of variable */ 
	  char *stmp1= strdup(rule->args[i]+2); /* allocate the rest of the string */ 
	  
	  if(!strncmp(type,"pp",2))       /* variable is past price */
	    { 
	      rule->argtype[i]  = "pp"; 
 	      rule->arglag[i] = my_atoi(stmp1); 
	      if (rule->arglag[i] == 0){
		fprintf(stderr,"Impossible lag: lag of price should be at least 1\n");
		exit(-2);
	      }
	      if((int) rule->arglag[i] > rule->maxlag)  /* update maximum rule lag */ 
 		rule->maxlag = rule->arglag[i]; 
	    }
	  else if(!strncmp(type,"ep",2))  /* variable is expected price */
	    { 
	      rule->argtype[i]  = "ep";
 	      rule->arglag[i]   = my_atoi(stmp1)+1;
	      if ((int) rule->arglag[i] > rule->maxlag) /* update maximum rule lag */ 
 		rule->maxlag = rule->arglag[i];
	    }	  
	  else if(!strncmp(type,"ap",2)) /* variable is average price */
	    { 
	      rule->argtype[i]  = "ap"; 
	      rule->arglag[i] = my_atoi(stmp1);  
	      if (rule->arglag[i] == 0){
		fprintf(stderr,"Impossible lag: lag of price should be at least 1\n");
		exit(-2);
	      }
	      if ((int) rule->arglag[i] > rule->maxlag) /* update maximum rule lag */ 
 		rule->maxlag = rule->arglag[i]; 
	    }
	  else if(!strncmp(type,"fp",2)) /* variable is fundamental price */
	    rule->argtype[i]  = "fp";
	  else
	    {
	      fprintf(stderr,"Wrong variable name: %s; exiting...\n",rule->args[i]); 
	      exit(-1);
	    }	  	  
	}
	break;
	
      case 1: /* initial share of the rule */
	rule->n = atof(value);
	if(rule->n<0){
	  fprintf(stderr,"Error: negative initial share; exiting...\n"); 
	  exit(-1);
	}	  		       
	break;
	
      default:/* Unknown suboption. */ 
	fprintf(stderr,"Unknown suboption: %s\n",value);  
	break; 
      }
      break;
    } 
  free(stmp1);
}


/* config file parsing */
/* ------------------- */
void parse_file(Rule **rule,int *rulesnum,const int maxtime,FILE *input) 
{ 
  Rule *R = *rule; 
  int Rnum=0; 
  
  char *line=NULL; 
  size_t linedim=0;
  
  /* parse line for rules specification */
  while(getline (&line,&linedim,stdin) != -1){ 
    
    /* ignore comments and empty lines */ 
    if (line[0] == '\n' || line[0] == '#' || line[0] == '\0') 
      continue; 
    
    /* possibly remove end-of-line character */ 
    if(line[strlen(line)-1] == '\n') line[strlen(line)-1]='\0'; 
    
    Rnum++; 
    R =(Rule *) my_realloc((void *) R, Rnum*sizeof(Rule)); 
     
    /* updating values from file */ 
    rule_from_string(&(R[Rnum-1]),line,maxtime); 
  } 
  
  *rulesnum=Rnum; 
  *rule=R; 
} 


/* initialization of a rule */ 
/* ------------------------ */
void initialize_predictors(Rule *rule, const Market market){
  
  int i,j;
  double sum;
  
  /* allocate space for values */
  rule->values = (double *) my_alloc(rule->argnum * sizeof(double));
  
  for(i=0;i<rule->argnum;i++){
    if(rule->argtype[i] == "pp")         /* lagged price */
      rule->values[i] = market.p[rule->maxlag - rule->arglag[i]];
    
    else if(rule->argtype[i] == "ep")    /* lagged expected price */
      rule->values[i] = market.p[rule->maxlag - rule->arglag[i]];
    
    else if(rule->argtype[i] == "ap")    /* average of past prices */
      {
	sum = 0;
       	for(j=0;j<rule->maxlag;j++)
	  sum = sum + market.p[j];
	rule->values[i] = sum/rule->maxlag;
      }
    else if(rule->argtype[i] == "fp")   /* fundamental price */
      rule->values[i] =  market.pf;
  }
  rule->pexp[rule->maxlag] = evaluator_evaluate(rule->r,rule->argnum,rule->args,rule->values);
}


/* computation of the predictions */ 
/* ------------------------------ */
void decide_expected_price(Rule *rule, const Market market)
{
  int i;
  
  for(i=0;i<rule->argnum;i++){
    if(rule->argtype[i] == "pp")         /* lagged price */
      {
	const unsigned lag = (int) rule->arglag[i];
	rule->values[i] = market.p[market.t-lag];	
      }
    else if(rule->argtype[i] == "ep")    /* lagged expected price */
      {
    	const unsigned lag =  (int) rule->arglag[i];
	if(market.t-lag>rule->maxlag-1)
	  rule->values[i] =  rule->pexp[market.t-lag];
	else rule->values[i] = market.p[market.t-lag-1];	
      }   
    else if(rule->argtype[i] == "ap")    /* average of past prices */
      {
	rule->values[i] = (rule->values[i]*(market.t-1)+market.p[market.t-1])/market.t;
      }
    else if(rule->argtype[i] == "fp")    /* fundamental price */
      {
    	rule->values[i] =  market.pf;
      }
    
  }
  
  rule->pexp[market.t] = evaluator_evaluate(rule->r,rule->argnum,rule->args,rule->values);
} 

/* performance measure is computed */ 
/* ------------------------------- */
double evaluate_performance(Rule *rule, const Market market)
{
  rule->u = market.eta * rule->u - 
    (market.p[market.t] - rule->pexp[market.t-1])*
    (market.p[market.t] - rule->pexp[market.t-1]);
  return rule->u;
}


/* output system state */ 
/* ------------------- */

void print_status(Rule * const rule,const int rulesnum, 
		  const Market market,const char *output)
{
  const unsigned t = market.t;
  const double p = market.p[t];
  int i;

  for(i=0;i<strlen(output);i++){
    switch(output[i]){
    case 't':
      printf(" %d ",t);
      break;
    case 'p' :
      printf(" %f ",p);
      break;   
    case 'y' :    /* dividend */
      printf(" %f ",market.y);
      break; 
    case 'n' :    /* share of robot traders */
      if (t>=market.pinum)
	printf(" %f ",market.nrobot);
      else printf("          ");
      break;   
    case 'a' :    /* average prediction */
      {
	int j;
	double sum;
	if (t>=market.pinum){
	  sum=0;
	  for (j=0;j<rulesnum;j++)
	    sum += rule[j].pexp[t]*rule[j].n;
	  printf(" %f ",sum);
	}
	else printf("          ");
      }
      break;
    case 'N' :    /* share of different rules */
      {
	int j; 
	for (j=0;j<rulesnum;j++)
	  if(t>=market.pinum)
	    printf(" %f ",(1.-market.nrobot)*rule[j].n);
	  else printf("          ");
      }
      break;   
    case 'F' :    /* forecasting price by rules */
      {
	int j; 
	for (j=0;j<rulesnum;j++)
	  if (rule[j].maxlag>t)
	    printf("          ");
	  else  printf(" %f ",rule[j].pexp[t]);
      }
      break;
    case 'P' :    /* actual prediction of the rule for the current period */
      {
	int j; 
	for (j=0;j<rulesnum;j++)
	  if (rule[j].maxlag>t-1)
	    printf("           ");
	  else  printf(" %f ",rule[j].pexp[t-1]);
      }
      break;
    case 'E' :    /* forecasting error of rules */
      {
	int j; 
	for (j=0;j<rulesnum;j++)
	  if (rule[j].maxlag+1>t)
	    printf("          ");
	  else printf(" %f ",p-rule[j].pexp[t-1]);
      }
      break;
    case 'U' :    /* performance of rules */
      {
	int j; 
	for (j=0;j<rulesnum;j++)
	  if (rule[j].maxlag+1>t)
	    printf("          ");
	  else printf(" %f ",fabs(rule[j].u));
      }
      break;
    }
  }
  printf("\n");
}


/* ------------------------------------ */
/* -------------   MAIN   ------------- */   
/* ------------------------------------ */
int main(int argc,char* argv[]){
  
  Rule *rule=NULL;
  int rulesnum=0;
  
  Market market;
  unsigned t; 
  
  /* dividend generator*/
  double (*dividend) (const Market market );

  /* READING COMMAND LINE */
  /* -------------------- */
  char opt;
  extern char *optarg;
  extern int optind, opterr, optopt;
  
  /* DEFAULT VALUES */
  /* ============== */
  
  /* simulation length */
  unsigned T=50;
  unsigned T_trans=0;
  unsigned T_skip=1;
  
  /* random seed */
  unsigned long int seed=0;
  
  /* robot traders exist? */
  unsigned robots = 0;  
  
  /* output type */
  char * output=strdup("tp");
  
  /* verbosity level */
  int o_verbose = 0;  
  
  /* the time at which price starts to be calculated */ 
  int max_t=0;  
  
  /* interest rate and dividend */
  market.rf = .05;  
  market.dividend_type = 0;
  market.process_mean  = 3;
  market.process_stdev = 0;
  
  /* default values of agents' parameters */
  market.beta = 1;
  market.eta = 0;
  market.delta = 0;
  
  /* initial share of robots*/
  market.nrobot = 0;    
  
  /* initial price */
  market.pinum = 1;
  market.pinit = (double *) my_alloc(sizeof(double));
  market.pinit[0] = 50;
  
  /* random number generator */
  market.rng = gsl_rng_alloc (gsl_rng_default);
  
  while((opt=getopt(argc,argv,"t:T:s:r:S:Y:P:b:e:d:O:fvh"))!=EOF){
    if(opt=='?'){
      fprintf(stderr,"option %c not recognized\n",optopt);
      return(-1);
    }

    else if(opt=='h'){
      /* print help */
      fprintf(stderr,"Model for explaining the exerimental results.\n");
      fprintf(stderr,"\nUsage: %s [options] < [config file]\n\n",argv[0]);
      fprintf(stderr,"options:\n");
      fprintf(stderr,"-t  simulation length                                      [50]\n");
      fprintf(stderr,"-T  transient length                                        [0]\n");
      fprintf(stderr,"-s  skip this number of steps in output                     [1]\n");
      fprintf(stderr,"-r  riskless return                                       [.05]\n");
      fprintf(stderr,"-S  seed of the random number generator                     [0]\n");
      fprintf(stderr,"-Y  dividend structure. Define parameters values using pairs   \n");
      fprintf(stderr,"    name=values. Names are 'type', 'mean', 'stdev'. Types are: \n");
      fprintf(stderr,"     0: normally distributed dividend                          \n");
      fprintf(stderr,"     1: real noise from the experiment  [type=0,mean=3,stdev=0]\n");
      fprintf(stderr,"-P  initial price history given as p0,p1,p2,p3,etc         [50]\n"); 
      fprintf(stderr,"-b  parameter 'beta', intensity of choice                   [1]\n");
      fprintf(stderr,"-e  parameter 'eta', memory in the performance measure      [0]\n");
      fprintf(stderr,"-d  parameter 'delta', fraction of conservative agents      [0]\n");
      fprintf(stderr,"-O  set the output, list of                                [tp]\n");
      fprintf(stderr,"     t:step, p:price, y:dividend, n:share of robots,           \n");
      fprintf(stderr,"     a:average prediction, P:prediction of the rules,          \n");     
      fprintf(stderr,"     N:share of the rules, U:performance of the rules,         \n");
      fprintf(stderr,"     F:price forecasts, E:forecasting errors                   \n");
      fprintf(stderr,"-f  existence of robot traders                              [0]\n");
      fprintf(stderr,"-v  verbose output                                             \n");
      fprintf(stderr,"-h  this help                                                  \n");
      return(0);
    }
       
    else if(opt=='t'){         /* set the simulation length */
      T=my_atoi(optarg);
    }
    
    else if(opt=='T'){         /* set the transient length */
      T_trans=my_atoi(optarg);
    }
     
    else if(opt=='s'){         /* set the step skipped in output */
      T_skip=my_atoi(optarg);
    }
    
    else if(opt=='r'){         /* set the riskless return */
      market.rf=atof(optarg);
    }    
    
    else if(opt=='S'){         /* set the rng seed */
      seed = strtoul(optarg,NULL,10);
    }
     
    else if(opt=='Y'){         /* set the process (yield or dividend) structure */
      
      char *const dividend_parameters[] = {"type","mean","stdev",NULL};
      int subopt;
      
      char *string=strdup(optarg);
      char *stmp1=string;
      char *subvalue;
      
      while  (*string != '\0')
	switch (subopt = getsubopt(&string, dividend_parameters, &subvalue)){
      	case 0:  /* type of process*/
	  market.dividend_type = my_atoi(subvalue);
	  break;
	case 1:  /* mean */
	  market.process_mean = atof(subvalue);
	  break;
	case 2:	 /* standard deviation */
	  market.process_stdev = atof(subvalue);
	  break;
	default: /* Unknown suboption.  */
	  fprintf(stderr,"Unknown parameter under -Y option\n");
	  exit(-1);
	}
      free(stmp1);
      free(string);
    }
    
    else if(opt=='P'){         /* set the initial price history */
      
      int i=0;
      char *string=strdup(optarg);
      char *stmp1=string;
      
      while (stmp1 != NULL){
	char *stmp2=strsep(&stmp1,",");
	market.pinit = (double *) my_realloc ((void *) market.pinit, (i+1)*sizeof(double));
	market.pinit[i]=atof(stmp2);
	i++;
      }
      
      market.pinum = i;
      free(stmp1); 
      free(string);
    }
    
    else if(opt=='b'){         /* set the "beta" parameter: intensity of choice */
      market.beta = atof(optarg);
    }  
    
    else if(opt=='e'){         /* set the "eta" parameter: memory in the performance measure */
      market.eta = atof(optarg);
    }
    
    else if(opt=='d'){         /* set the "delta" parameter: fraction of conservative agents */
      market.delta = atof(optarg);
    }
    
    else if(opt=='O'){        /* set the output type*/
      
      int i = strspn(optarg,"tpynaPNUFE");
      
      if( i != strlen(optarg) ){    /* check if variables are meaningful */
	fprintf (stderr,"Unknown output type %c\n", optarg[i]);
	exit(2);
      }
      
      free(output);           /* remove default and define new */
      output = strdup(optarg);      
    }
    
    else if(opt=='f'){         /* set whether the robot traders exist */
      robots = 1;
    }   
    
    else if(opt=='v'){         /* verbosity of output */
      o_verbose=1;
    }
  }
  
  if(T<market.pinum) 
    T = market.pinum;
  /* END OF COMMAND LINE PROCESSING */
  
  /* initialize the random number generator */
  gsl_rng_set(market.rng,seed);

  /* parse input for rules specification */
  parse_file(&rule,&rulesnum,T+T_trans+1,stdin);
 
  {/* computation of the period when all the rules can be evaluated */
    int i;
    for(i=0;i<rulesnum;i++){
      if(rule[i].maxlag > rule[i].t)
	rule[i].t = rule[i].maxlag;
      if(rule[i].t > max_t)
	max_t = rule[i].t;
    }
    if(max_t>market.pinum){
      fprintf(stderr,"The initial history is too short. At least %d prices should be provided.\n",max_t);
      exit(-1);
    }
  }
  
  /* initialize past prices */
  { 
    int i;
    
    market.p = (double *) my_alloc((T+T_trans+1) * sizeof(double));
    for(i=0;i<market.pinum;i++)
      market.p[i] = market.pinit[i];
  }
  
  /* set dividend structure */ 
  switch(market.dividend_type){
  case 0:  /* CASE 0: dividend is normally distributed */
    dividend = dividend_normal;
    market.pf= market.process_mean/market.rf;
    break;
    
  case 1:  /* CASE 1: dividend is taken from the experiment */
    dividend = dividend_experiment; 
    market.pf= market.process_mean/market.rf;
    break;
    
  default:
    fprintf (stderr,"Unknown process type %d\n",market.dividend_type);
    exit(2);
  }
  
  { /* agent predictors initialization */
    int i;
    for(i=0;i<rulesnum;i++)
      initialize_predictors(&(rule[i]),market);
  }
  
  { /* normalization of initial fractions */
    double sum=0;
    int i;
    for(i=0;i<rulesnum;i++)
      sum = sum + rule[i].n;
    for(i=0;i<rulesnum;i++)
      if(sum==0)
	rule[i].n=1./rulesnum;
      else rule[i].n=rule[i].n/sum;
  }
  
  /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
  if(o_verbose) {  /* check initialization phase */
    int i,j;
    
    for(i=0;i<rulesnum;i++){
      char * string = evaluator_get_string (rule[i].r);
      fprintf (stderr,"Rule %d:  ",i);
      fprintf (stderr," r=%s\n",string);   
      fprintf (stderr," -- initial time to be used=%d\n",rule[i].t); 
      fprintf (stderr," -- initial fraction assigned to the rule=%f\n",rule[i].n);
      fprintf (stderr," Arguments:  type      lag    value  \n");
      for(j=0;j<rule[i].argnum;j++)
	fprintf (stderr,"  %d           %s        %d      %f   \n",j,rule[i].argtype[j],rule[i].arglag[j],rule[i].values[j]);
      fprintf (stderr," ---- At t=%d this rule gives prediction %f for price at time %d\n",rule[i].maxlag,rule[i].pexp[rule[i].maxlag],rule[i].maxlag+1);
    }
    
    fprintf(stderr,"Market:\n");
    fprintf(stderr,"  riskless interest rate = %f\n",market.rf);
    switch(market.dividend_type){
    case 0:
      fprintf(stderr,"  dividend type          = normal dividend\n");
      fprintf(stderr,"  dividend mean          = %f\n",market.process_mean);
      fprintf(stderr,"  dividend stdev         = %f\n",market.process_stdev);
      break;
    case 1:
      fprintf(stderr,"  dividend type          = taken from experiment\n");
      fprintf(stderr,"  dividend mean          = %f\n",market.process_mean);
      fprintf(stderr,"  dividend stdev         = %f\n",market.process_stdev);
      break;
    } 
    fprintf(stderr,"  fundamental price      = %f\n",market.pf);
    fprintf(stderr," Initial price history:\n");
    for(i=0;i<market.pinum;i++)
      fprintf(stderr," %f",market.p[i]); 
    fprintf(stderr, "\n");
    fprintf(stderr,"  intensity of choice, beta                = %f\n",market.beta);
    fprintf(stderr,"  memory in the performance measure, eta   = %f\n",market.eta);
    fprintf(stderr,"  fractions of conservative players, delta = %f\n",market.delta);
    fprintf(stderr,"    robots ARE %sin the market\n", (robots == 0) ? "NOT " : "");
    fprintf(stderr, "\n");
    
    fprintf(stderr,"Simulation:\n");
    fprintf(stderr," total length           = %d\n",T+T_trans);
    fprintf(stderr," transient length       = %d\n",T_trans);
    fprintf(stderr," rng seed               = %ld\n",seed);
    
    fprintf(stderr,"End of an initial check. Simulation can start!\n\n");
  }
  /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
  
  /* === START MAIN LOOP === */
  for (t=1;t<=T+T_trans;t++){
    int i;
    double price, sum;
    
    /* update time index -------------------------- */
    market.t=t;
    
    /* update the forecasts of different rules --- */
    for (i=0;i<rulesnum;i++)
      /* if (t < rule[i].maxlag) the forecast is not yet available, */
      /* if (t = rule[i].maxlag) the forecast was only initialized  */
      if(t > rule[i].maxlag)
	decide_expected_price(&(rule[i]),market);
    
    /* generate new yield and dividend ------------ */
    market.y=dividend(market);
    
    /* define new price --------------------------- */
    if(t >= market.pinum){
      price=0;
      for(i=0;i<rulesnum;i++)   /* when price is endogenous, all the rules are in use */
	price += rule[i].n*rule[i].pexp[t];
      market.p[t] = (market.nrobot*market.pf + (1.-market.nrobot)*price+market.y)/(1.+market.rf); 
    }
    
    /* evaluate forecasting rules ----------------- */
    for (i=0;i<rulesnum;i++)
      if(t >= max_t+1)   /* evaluation of all rules starts simultaneously */
	evaluate_performance(&(rule[i]),market);
    
    /* print output -------------------------------- */
    if(t>T_trans && t%T_skip == 0)
      print_status(rule,rulesnum,market,output);
    
    /* update shares of robots  -------------- */
    if(t >= max_t && robots != 0)
      market.nrobot = 1 - exp(-fabs(market.p[t]-market.pf)/200);
    
    /* update shares of rules ---------------- */
    if(t >= max_t+1){
      sum = 0;
      for (i=0;i<rulesnum;i++)
	sum += exp(rule[i].u*market.beta);
      for (i=0;i<rulesnum;i++)
	rule[i].n = market.delta*rule[i].n + (1.-market.delta)*exp(rule[i].u*market.beta)/sum; 
    }
    
    /* test if floating point exception has been raised */
    testfpe();
  }
  /* === END MAIN LOOP === */
  
  return 0;
}


